from django.apps import AppConfig


class ExporterConfig(AppConfig):
    name = "django_infra.exporter"

    def ready(self):
        # load receivers
        # import django_infra.exporter.receivers  # noqa F401
        ...



import json
import logging
import os
import subprocess
import typing

logger = logging.getLogger(__file__)


def get_failure_msg(key, default, var, validation, allow_none):
    return "\n".join(
        [
            f"Failed to load env var with key:{key},",
            f" default:{default}, ",
            f"loaded val: {var}",
            f"validation:{validation} ",
            f"allow_none:{allow_none}",
        ]
    )


def load_env_val(
    key: str, default=None, allow_none=False, validation: typing.Callable = None
) -> typing.Any:
    """

    Parameters
    ----------
    key
    default: default value if not defined
    allow_none: is it ok for val not to be defined
    validation: validation function for loaded variable, e.g. lambda x: isinstance(x,dict)

    Returns
    -------

    """
    var = os.environ.get(key, default=default)
    if var is not None and isinstance(var, str):
        if var.startswith("_json_"):
            try:
                var = json.loads(var[6:])
            except json.JSONDecodeError:
                raise RuntimeError(
                    get_failure_msg(key, default, var, validation, allow_none)
                )
    elif var is None and not allow_none:
        raise RuntimeError(get_failure_msg(key, default, var, validation, allow_none))
    if validation and not validation(var):
        raise RuntimeError(
            f"Validation error: "
            f"{get_failure_msg(key, default, var, validation, allow_none)}"
        )
    return var


def run_command(
    command: str | typing.List[str], env: dict = None, background=False
) -> subprocess.Popen[str]:
    """
    Execute a command with output in a fixed-height window.
    Prints all output only on failure, otherwise remains silent.
    Raises CalledProcessError on failure with last output lines.
    """
    if isinstance(command, str):
        command = command.split(" ")

    command_style = "\033[38;5;45m"
    exec_command = " ".join(list(map(str, command)))

    logger.info(f"[ {command_style} EXECUTING: {exec_command}\033[0m ]")

    # Configure process parameters based on background flag
    stdout = subprocess.PIPE
    stderr = subprocess.STDOUT
    if background:
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL

    # Initialize process once
    proc = subprocess.Popen(
        command,
        stdout=stdout,
        stderr=stderr,
        text=True,
        bufsize=1,
        env=env,
    )

    # For background processes, don't wait
    if background:
        return proc

    # For foreground processes, capture output and check return code
    stdout_data, _ = proc.communicate()
    output_lines = stdout_data.splitlines() if stdout_data else []

    # If command failed, print all captured output and raise exception
    if proc.returncode != 0:
        for line in output_lines:
            print(line)
        raise subprocess.CalledProcessError(
            proc.returncode, command, output=stdout_data
        )
    return proc



from django.db import models

from django_infra.db.models import UpdatableModel


class FeatureFlagManager(models.Manager):
    def create(self, *args, **kwargs):
        # Prevent direct creation of objects
        raise NotImplementedError(
            "Direct creation of FeatureFlag objects is not allowed."
        )

    def delete(self, *args, **kwargs):
        # Prevent deletion of objects
        raise NotImplementedError("Deletion of FeatureFlag objects is not allowed.")


class FeatureFlag(UpdatableModel):
    id = models.CharField(
        max_length=50,
        unique=True,
        primary_key=True,
    )
    active = models.BooleanField(default=False)
    value = models.IntegerField(default=0)
    value_str = models.CharField(default="", max_length=100)
    # Assign the custom manager
    objects = FeatureFlagManager()

    def __bool__(self):
        return self.active



from django.apps import AppConfig


class FeatureFlagConfig(AppConfig):
    name = "django_infra.feature_flags"

    def ready(self):
        # load receivers
        import django_infra.feature_flags.receivers  # noqa F401



from .flags import retrieve_feature_flag_from_db


def feature_flags_cache_clear_middleware(get_response):
    """Clear retrieved feature-flag cache on per-request bases to avoid db spam.

    Given that the same feature flag might be invoked multiple times in the same
    request, we avoid db spam by clearing lru cache on flag retrieval.
    """

    def middleware(request):
        retrieve_feature_flag_from_db.cache_clear()
        return get_response(request)

    return middleware



from django.contrib import admin

from .models import QueryExport


@admin.register(QueryExport)
class ExporterAdmin(admin.ModelAdmin):
    list_display = ("id",)
    list_filter = ("state",)
    search_fields = ("id",)



import dataclasses
import time
from functools import cached_property

from django.db import models
from django.db.models import TextChoices

from django_infra.db.models import UpdatableModel


class QueryExportManager(models.Manager): ...


class ExportState(TextChoices):
    SCHEDULED = "scheduled", "Scheduled"
    PROCESSING = "processing", "Processing"
    SUCCESS = "success", "Success"
    FAIL = "fail", "Fail"


@dataclasses.dataclass
class ExportMetadata:
    start_time: float = dataclasses.field(default_factory=time.time)
    file_size: int = 0
    job_time: float = 0.0
    progress_percent: int = 0
    row_count: int = 0
    error_log: str = ""

    def update_progress(self, processed: int, total: int):
        self.row_count = processed
        self.progress_percent = int(processed * 100 / total) if total else 0

    def finalize(self, file_size: int, processed: int):
        self.job_time = time.time() - self.start_time
        self.file_size = file_size
        self.row_count = processed
        self.progress_percent = 100

    @property
    def data(self) -> dict:
        return dataclasses.asdict(self)


class QueryExport(UpdatableModel):

    id = models.CharField(
        max_length=64,
        unique=True,
        primary_key=True,
    )
    state = models.CharField(max_length=16, choices=ExportState.choices)

    metadata = models.JSONField()

    file = models.FileField()
    format = models.CharField(
        max_length=16,
    )
    # Assign the custom manager
    objects = QueryExportManager()

    @cached_property
    def export_metadata(self) -> ExportMetadata:
        return ExportMetadata(**self.metadata)



from __future__ import annotations

import dataclasses
from functools import lru_cache
from typing import TYPE_CHECKING, Set

from django_infra.env import load_env_val

if TYPE_CHECKING:
    from django_infra.feature_flags.models import FeatureFlag


@lru_cache  # this is popped in a post save signal on FeatureFlag.
def retrieve_feature_flag_from_db(*, key: str, active=False) -> "FeatureFlag":
    from django_infra.feature_flags.models import FeatureFlag

    """Get feature flag value from db, Ensure sync is run once.
    Property overriding every attribute in FeatureFlags, set by FeatureFlagsMeta"""
    # set a boolean flag stating we have synced with the database on any initial use.
    ff, _ = FeatureFlag.objects.get_or_create(pk=key, defaults=dict(active=active))
    return ff


def register_feature_flag(flag_name, default=False) -> "FeatureFlag":
    return_val = property(
        fget=lambda self: retrieve_feature_flag_from_db(
            key=flag_name, active=load_env_val(flag_name, default=default)
        )
    )
    return return_val


@dataclasses.dataclass
class FeatureFlags:
    """Exposed database stored flags as bool properties efficiently.
    A flag is loaded once request
    Example:
        >>> # settings.py:
        >>> Flags = FeatureFlags()
        >>> Flags.DEFAULT_FLAG=register_feature_flag('DEFAULT_FLAG',default=True),
        >>>
        >>> # somewhere else.
        >>> from django_infra.feature_f1lags import FeatureFlags # noqa
        >>> def my_function(*args,**kwargs):
        >>>     if FeatureFlags.DEFAULT_FLAG:
        >>>         print("default flag is enable")
        >>>         FeatureFlags.DEFAULT_FLAG.update(active=False)
        >>>         print("default flag is disabled")
        >>>     return args
    """

    _flags: set = dataclasses.field(init=False, default_factory=set)

    def __setattr__(self, key, value):
        if key not in {"_flags"}:
            # some minimal validation that register_feature_flag was used for setup.
            if not isinstance(value, property):
                raise ValueError(
                    f"Flag value must be created via {register_feature_flag}"
                )
            setattr(self.__class__, key, value)
            self._flags.add(key)
            return

        super().__setattr__(key, value)

    @property
    def choices(self) -> Set[str]:
        return self._flags

    def __hash__(self):
        return 10


FeatureFlags = FeatureFlags()
flags = FeatureFlags



import collections
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional

from django.conf import settings
from django.db.models import ForeignKey, ManyToManyField, Prefetch, QuerySet
from djangorestframework_camel_case.util import underscoreize
from drf_partial_response import constants
from drf_partial_response.views import OptimizedQuerySetBase, OptimizedQuerySetMixin
from jsonmask import parse_fields
from rest_framework import exceptions


@dataclass
class Mutation:
    _mutations: Optional[List[Callable[[QuerySet], QuerySet]]] = None

    priority: int = 0
    base_queryset: QuerySet = None

    @property
    def mutations(self) -> List[Callable[[QuerySet], QuerySet]]:
        if self._mutations is None:
            self._mutations = list()
        return self._mutations

    def add(
        self,
        mutation: Callable[[QuerySet], QuerySet],
        base_queryset: QuerySet,
        priority: int = 0,
    ):
        if priority > self.priority or self.base_queryset is None:
            self.base_queryset = base_queryset
        self.mutations.append(mutation)

    def apply_all(self) -> QuerySet:
        queryset = self.base_queryset
        for mutation in self.mutations:
            queryset = mutation(queryset)
        return queryset


class PrefetchKwargs:
    def __init__(self, lookup: str, to_attr=None):
        self.lookup = lookup
        self.to_attr = to_attr

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(f"{self.lookup}_{self.to_attr}")


class OptimizedQuerySetAnnotationsMixin(
    OptimizedQuerySetMixin, metaclass=OptimizedQuerySetBase
):
    """Extends OptimizedQuerySetMixin to allow for combining annotations rather than
    prefetching at each data predicate instantiation.
    by using add_nested_queryset_mutation, all mutations will be stored and applied once
    optimization is called at the end.

    Optionally pre-fetches all relations associated with the model via
    """

    _mutations: Dict[PrefetchKwargs, Mutation] = None
    # default fields to default to None defaults to all, empty list will default to None.
    default_requested_fields_to: None | List[str] = None
    auto_prefetch_related = False

    @property
    def requested_fields(self):
        # have to do this manually instead of calling super() due
        # to cached_property breaking in drf_partial...
        # also allows us to set fields directly into _requested_fields
        if not hasattr(self, "_requested_fields"):
            fields_name = getattr(
                settings, "REST_FRAMEWORK_JSONMASK_FIELDS_NAME", constants.FIELDS_NAME
            )
            self._requested_fields = underscoreize(
                parse_fields(self.request.query_params.get(fields_name))
            )
        if (
            isinstance(self.default_requested_fields_to, list)
            and self._requested_fields is None
        ):
            self._requested_fields = {
                field: dict() for field in self.default_requested_fields_to
            }
        return self._requested_fields

    @property
    def requested_fields_set(self):
        fields = self.requested_fields or dict()
        return set(fields.keys())

    def get_serializer_context(self):
        """Patch over original get serializer context.
        avoids accessing GET & simplifies.
        """
        context = super(OptimizedQuerySetMixin, self).get_serializer_context()

        if self.requested_fields and self.excluded_fields:
            raise exceptions.ParseError(
                detail="Cannot provide both requested and excluded fields"
            )
        if self.requested_fields:
            context["requested_fields"] = self.requested_fields
        if self.excluded_fields:
            context["excluded_fields"] = self.excluded_fields
        return context

    @property
    def mutations(self) -> Dict[PrefetchKwargs, Mutation]:
        if self._mutations is None:
            self._mutations = collections.defaultdict(Mutation)
        return self._mutations

    def _clean(self):
        self._mutations = None

    def get_queryset(self):
        self._clean()
        queryset = super(OptimizedQuerySetMixin, self).get_queryset()
        queryset = self.optimize_queryset(queryset)
        if "Auto-Prefetch" in self.request.headers:
            self.prefetch_related_objects(queryset)
        queryset = self.apply_all_mutations(queryset)
        return queryset

    def add_nested_queryset_mutation(
        self,
        lookup: str,
        base_queryset: QuerySet,
        queryset_transformer_lambda: Callable[[QuerySet], QuerySet] = None,
        base_queryset_priority: int = 0,
        to_attr: str = None,
    ):
        """Only for nested lookups.
        >>> # This will fail (initial annotation will be lost)
        >>> my_queryset.prefetch_related('my_attr',queryset=my_queryset_annotated_1)
        >>> my_queryset.prefetch_related('my_attr',queryset=my_queryset_annotated_2)
        >>> # This will pass:
        >>> self.add_nested_queryset_mutation('my_attr',queryset=my_base_queryset,queryset_transformer_lambda=lambda q:q.annotated_1())
        >>> self.add_nested_queryset_mutation('my_attr',queryset=my_base_queryset,queryset_transformer_lambda=lambda q:q.annotated_2())
        Parameters
        ----------
        lookup
        base_queryset
        queryset_transformer_lambda
        base_queryset_priority
        to_attr
        Returns
        -------

        """
        if queryset_transformer_lambda is None:
            queryset_transformer_lambda = lambda q: q.all()  # noqa
        self.mutations[PrefetchKwargs(lookup=lookup, to_attr=to_attr)].add(
            queryset_transformer_lambda, base_queryset, base_queryset_priority
        )

    def prefetch_related_objects(self, queryset):
        # Retrieve the model class associated with this queryset
        model = queryset.model
        # Get all the fields of the model
        fields = [
            field
            for field in model._meta.get_fields()
            if isinstance(field, (ForeignKey, ManyToManyField))
        ]
        for field in fields:
            self.add_nested_queryset_mutation(
                field.name, field.related_model.objects.all()
            )

    def apply_all_mutations(self, queryset):
        for prefetch_kwargs, mutation in self.mutations.items():
            queryset = queryset.prefetch_related(
                Prefetch(
                    lookup=prefetch_kwargs.lookup,
                    queryset=mutation.apply_all(),
                    to_attr=prefetch_kwargs.to_attr,
                )
            )
        return queryset


class CaseInsensitiveEnumMixin:
    _set = None  # Initialize the attribute to store the set

    @classmethod
    def _initialize_set(cls):
        # This creates a set with both standard and lowercase values
        if cls._set is None:
            cls._set = {e.name for e in cls} | {e.name.lower() for e in cls}

    @classmethod
    def __contains__(cls, item):
        cls._initialize_set()  # Ensure the set is initialized
        # Perform a case-insensitive check
        return item in cls._set or item.lower() in cls._set



from rest_framework.serializers import ModelSerializer

from django_infra.api import filters
from django_infra.api.views import FilteredPartialResponseModelViewSet
from django_infra.exporter.models import QueryExport


class QueryExportSerializer(ModelSerializer):
    class Meta:
        model = QueryExport
        fields = "__all__"
        read_only_fields = ("state", "file", "format", "id", "metadata")


class QueryExportViewSet(FilteredPartialResponseModelViewSet):
    model = QueryExport
    serializer_class = QueryExportSerializer
    permission_classes = []
    filters = [
        filters.OrderingFilter(
            name="sort",
            fields=[
                filters.FilterField(internal="id"),
            ],
        ),
        filters.ExactMatchFilter(
            fields=[filters.TypedFilterField(type_=str, internal="state")]
        ),
        filters.IContainsFilter(
            fields=[
                filters.MultiInternalFilterField(internal=["file"], external="search")
            ]
        ),
    ]



import pytest
from django.conf import settings
from django.db import transaction


def disable_on_transaction_commit():
    """
    Break this out to allow for easier patching. Basically if you're test is transactional then
    you'll want to patch this to False to sure that the behaviour matches the release code
    """
    return settings.IS_TEST


def on_transaction_commit(func):
    """Decorator to delay function execution until after the current transaction commits."""

    def wrapper(*args, **kwargs):
        # Use a lambda to capture args and kwargs and pass them to the function
        if disable_on_transaction_commit():
            func(*args, **kwargs)
        else:
            transaction.on_commit(lambda: func(*args, **kwargs))

    return wrapper


def synchronous_in_test(func):
    def wrapper(*args, **kwargs):
        existing_state = settings.CELERY_TASK_ALWAYS_EAGER
        if settings.IS_TEST:
            settings.CELERY_TASK_ALWAYS_EAGER = True
        try:
            result = func(*args, **kwargs)
        except Exception:
            raise
        finally:
            if settings.IS_TEST:
                settings.CELERY_TASK_ALWAYS_EAGER = existing_state
        return result

    return wrapper


def integration_test(
    func,
):
    return pytest.mark.skipif(not settings.INTEGRATION_TEST, reason="integration test")(
        pytest.mark.integration(func)
    )


def qa_test(
    func,
):
    return pytest.mark.skipif(not settings.QA_TEST, reason="qa test")(
        pytest.mark.qa(func)
    )


def prod_test(
    func,
):
    return pytest.mark.skipif(not settings.PROD_TEST, reason="prod test")(
        pytest.mark.prod(func)
    )



from django.urls import include, path
from rest_framework.routers import DefaultRouter

from django_infra.feature_flags.views import FeatureFlagViewSet

app_name = "feature_flags"

# Create a router and register our viewsets with it, using an empty prefix.
router = DefaultRouter()
router.register(r"", FeatureFlagViewSet, basename="featureflag")

# The API URLs are now determined automatically by the router.
urlpatterns = [
    path("", include(router.urls)),
]



from rest_framework import serializers

from django_infra.feature_flags.models import FeatureFlag


class FeatureFlagSerializer(serializers.ModelSerializer):
    class Meta:
        fields = "__all__"
        model = FeatureFlag
